"""Core NIfTI-MRS class.
For more information on NIfTI-MRS see https://github.com/wtclarke/mrs_nifti_standard

Copyright William Clarke, University of Oxford, 2023
"""
import json
from pathlib import Path
import re

import nibabel as nib
import numpy as np

from fsl.data.image import Image

from . import validator
from .hdr_ext import Hdr_Ext
from .definitions import dimension_tags, standard_defined


class NIFTIMRS_DimDoesntExist(Exception):
    pass


class NotNIFTI_MRS(Exception):
    pass


class NIFTI_MRS():
    """A class to load and represent NIfTI-MRS formatted data.
    Utilises the fslpy Image class and nibabel nifti headers."""

    def __init__(self, *args, **kwargs):
        """Create a NIFTI_MRS object with the given image data or file name.

        Aguments mirror those of the leveraged fsl.data.image.IMage class.

        :arg image:      A string containing the name of an image file to load,
                         or a Path object pointing to an image file, or a
                         :mod:`numpy` array, or a :mod:`nibabel` image object,
                         or an ``Image`` object.

        :arg name:       A name for the image.

        :arg header:     If not ``None``, assumed to be a
                         :class:`nibabel.nifti1.Nifti1Header` or
                         :class:`nibabel.nifti2.Nifti2Header` to be used as the
                         image header. Not applied to images loaded from file,
                         or existing :mod:`nibabel` images.

        :arg xform:      A :math:`4\\times 4` affine transformation matrix
                         which transforms voxel coordinates into real world
                         coordinates. If not provided, and a ``header`` is
                         provided, the transformation in the header is used.
                         If neither a ``xform`` nor a ``header`` are provided,
                         an identity matrix is used. If both a ``xform`` and a
                         ``header`` are provided, the ``xform`` is used in
                         preference to the header transformation.

        :arg dataSource: If ``image`` is not a file name, this argument may be
                         used to specify the file from which the image was
                         loaded.

        :arg loadMeta:   If ``True``, any metadata contained in JSON sidecar
                         files is loaded and attached to this ``Image`` via
                         the :class:`.Meta` interface. if ``False``, metadata
                         can be loaded at a later stage via the
                         :func:`loadMeta` function. Defaults to ``False``.

        :arg dataMgr:    Object implementing the :class:`DataManager`
                         interface, for managing access to the image data.

        All other arguments are passed through to the ``nibabel.load`` function
        (if it is called).
        """
        # Handle various options for the first (data source) argument
        if isinstance(args[0], np.ndarray):
            # If generated from np.array include conjugation
            # to make sure generation from data of existing NIfTI-MRS
            # object results in consistent phase/freq convention.
            args = list(args)
            args[0] = args[0].conj()
            filename = None
        elif isinstance(args[0], Path):
            args = list(args)
            filename = args[0].name
            args[0] = str(args[0])
        elif isinstance(args[0], str):
            args = list(args)
            filename = Path(args[0]).name
        elif isinstance(args[0], NIFTI_MRS):
            args = list(args)
            filename = args[0].filename
            args[0] = args[0]._image

        # Instantiate Image object
        self._image = Image(*args, **kwargs)

        # Store original filename for reports etc
        self._filename = filename

        # Check that file meets minimum requirements
        try:
            if float(self.nifti_mrs_version) < 0.2:
                raise NotNIFTI_MRS('NIFTI-MRS > V0.2 required.')
        except IndexError:
            raise NotNIFTI_MRS('NIFTI-MRS intent code not set.')

        hdr_ext_codes = self.header.extensions.get_codes()
        if 44 not in hdr_ext_codes:
            raise NotNIFTI_MRS('NIFTI-MRS must have a header extension.')

        self._hdr_ext = Hdr_Ext.from_header_ext(
            json.loads(
                self.header.extensions[hdr_ext_codes.index(44)].get_content()),
            dimensions=self.ndim)

        try:
            self.nucleus
            self.spectrometer_frequency
        except KeyError:
            raise NotNIFTI_MRS('NIFTI-MRS header extension must have nucleus and spectrometerFrequency keys.')

    def __getitem__(self, sliceobj):
        '''Apply conjugation at use. This swaps from the
        NIFTI-MRS and Levvit inspired right-handed reference frame
        to a left-handed one, which FSL-MRS development started in.'''
        # print(f'getting {sliceobj} to conjugate {super().__getitem__(sliceobj)}')
        return self._image[sliceobj].conj()

    def __setitem__(self, sliceobj, values):
        '''Apply conjugation back at write. This swaps from the
        FSL-MRS left handed convention to the NIFTI-MRS and Levvit
        inspired right-handed reference frame.'''
        # print(f'setting {sliceobj} to conjugate of {values[0]}')
        # print(super().__getitem__(sliceobj)[0])
        self._image[sliceobj] = values.conj()
        # print(super().__getitem__(sliceobj)[0])

    # Implement useful calls to attributes of the image class object. Should I just be using inheretence here? Not sure.
    @property
    def header(self):
        """Returns NIfTI-MRS header object"""
        return self._image.header

    @property
    def shape(self):
        """Returns data shape"""
        return self._image.shape

    @property
    def ndim(self):
        """Returns number of data dimensions"""
        return self._image.ndim

    @property
    def nifti_mrs_version(self):
        """Get NIfTI-MRS version string."""
        tmp_vstr = self._image.header.get_intent()[2].split('_')
        return tmp_vstr[1].lstrip('v') + '.' + tmp_vstr[2]

    def set_version_info(self, major, minor):
        """Puts mrs_v{major}_{minor} into intent_name"""
        self.header['intent_name'] = f'mrs_v{major}_{minor}'.encode()

    @property
    def dwelltime(self):
        '''Return dwelltime in seconds'''
        return self.header['pixdim'][4]

    @dwelltime.setter
    def dwelltime(self, new_dt):
        """Sets new dwelltime (pixdim[4]). Units = seconds"""
        self.header['pixdim'][4] = new_dt

    @property
    def spectralwidth(self):
        '''Return spectral width in Hz'''
        return 1 / self.dwelltime

    @property
    def bandwidth(self):
        '''Alias for spectralwidth (Hz)'''
        return self.spectralwidth

    @property
    def nucleus(self):
        """Returns resonant nucleus string(s) - returns list"""
        return self.hdr_ext['ResonantNucleus']

    @property
    def spectrometer_frequency(self):
        '''Central or spectrometer frequency in MHz - returns list'''
        return self.hdr_ext['SpectrometerFrequency']

    def getAffine(self, *args):
        """Return an affine transformation which can be used to transform
        coordinates from ``from_`` to ``to``.

        Valid values for the ``from_`` and ``to`` arguments are:

         - ``'voxel'``: The voxel coordinate system
         - ``'world'``: The world coordinate system, as defined by the image
           sform/qform
         - ``'fsl'``: The FSL coordinate system (scaled voxels, with a
           left-right flip if the sform/qform has a positive determinant)

        :arg from_: Source coordinate system
        :arg to:    Destination coordinate system
        :returns:   A ``numpy`` array of shape ``(4, 4)``
        """
        return self._image.getAffine(*args)

    @property
    def hdr_ext(self):
        '''Return MRS JSON header extension object.'''
        return self._hdr_ext

    def _save_hdr_ext(self):
        """Method to place Hdr_ext object into underlying Image object"""
        extension = nib.nifti1.Nifti1Extension(
            44,
            self._hdr_ext.to_json().encode('UTF-8'))
        self.header.extensions.clear()
        self.header.extensions.append(extension)

    @hdr_ext.setter
    def hdr_ext(self, new_hdr):
        '''Update MRS JSON header extension from python dict or Hdr_Ext object'''
        if isinstance(new_hdr, dict):
            validator.validate_hdr_ext(json.dumps(new_hdr), self.ndim)
            self._hdr_ext = Hdr_Ext.from_header_ext(new_hdr, dimensions=self.ndim)
        elif isinstance(new_hdr, Hdr_Ext):
            validator.validate_hdr_ext(new_hdr.to_json(), self.ndim)
            self._hdr_ext = new_hdr
        else:
            raise TypeError('Passed header extension must be a dict or Hdr_Ext object')

        # Update the underlying Image object headers with new hdr extension
        self._save_hdr_ext()

    # Utility / legacy functions for hdr extension manipulation
    def add_hdr_field(self, key, value, doc=None):
        """Add a field to the header extension

        :param key: Field key
        :type key: str
        :param value: Value of field to add
        :param doc: Use to convey meaning of user-defined header value.
        :type doc: optional, str
        """
        dim_n = re.compile(r'dim_[567].*')
        if dim_n.match(key):
            raise ValueError('Modify dimension headers through dedicated methods.')

        new_hdr = self.hdr_ext
        if key in standard_defined:
            new_hdr.set_standard_def(key, value)
        else:
            if doc is None:
                raise ValueError('Please provide info about user defined value.')
            new_hdr.set_user_def(key=key, value=value, doc=doc)

        self.hdr_ext = new_hdr

    def remove_hdr_field(self, key):
        """Remove a field from the header extension

        :param key: Key to remove
        :type key: str
        """
        if key == 'SpectrometerFrequency' or key == 'ResonantNucleus':
            raise ValueError('You cannot remove the required metadata.')

        dim_n = re.compile(r'dim_[567].*')
        if dim_n.match(key):
            raise ValueError('Modify dimension headers through dedicated methods.')

        curr_hdr_ext = self.hdr_ext
        if key in standard_defined:
            curr_hdr_ext.remove_standard_def(key)
        else:
            curr_hdr_ext.remove_user_def(key)
        self.hdr_ext = curr_hdr_ext

    @property
    def filename(self):
        '''Name of file object was generated from.
        Returns empty string if N/A.'''
        if self._filename:
            return self._filename
        else:
            return ''

    @property
    def dim_tags(self):
        """Return the three higher dimension tags"""
        return self._read_dim_tags()

    def _read_dim_tags(self):
        """Read dim tags from current header extension"""
        dim_tags = [None, None, None]
        std_tags = ['DIM_COIL', 'DIM_DYN', 'DIM_INDIRECT_0']
        for idx in range(3):
            curr_dim = idx + 5
            curr_tag = f'dim_{curr_dim}'
            if curr_tag in self.hdr_ext:
                dim_tags[idx] = self.hdr_ext[curr_tag]
            elif curr_dim < self.ndim:
                dim_tags[idx] = std_tags[idx]
        return dim_tags

    def dim_position(self, dim_tag):
        '''Return position of dim if it exists.'''
        if dim_tag in self.dim_tags:
            return self._dim_tag_to_index(dim_tag)
        else:
            raise NIFTIMRS_DimDoesntExist(f"{dim_tag} doesn't exist in list of tags: {self.dim_tags}")

    def _dim_tag_to_index(self, dim):
        '''Convert DIM tag str or index (4, 5, 6) to numpy dimension index'''
        if isinstance(dim, str):
            if dim in self.dim_tags:
                dim = self.dim_tags.index(dim)
                dim += 4
        return dim

    def set_dim_tag(self, dim, tag, info=None, header=None):
        """Set or update the dim_N, dim_N_info, and dim_N_header fields

        Tag must be one of the standard-defined tags (e.g. DIM_DYN)

        :param dim: The existing dim tag or python dimension index (i.e. N-1)
        :type dim: str or int
        :param tag: New tag
        :type tag: str
        :param info: New info string
        :type info: str
        :param header: dict containing the dimension headers
        :type header: dict
        """
        if tag not in dimension_tags.keys():
            raise ValueError(f'Tag must be one of: {", ".join(list(dimension_tags.keys()))}.')

        dim = self._dim_tag_to_index(dim)

        if header is not None:
            # Check size
            def size_chk(obj):
                # Allow for expansion along the next dimension
                if dim == self.ndim:
                    dim_len = 1
                else:
                    dim_len = self.shape[dim]
                if len(obj) != dim_len:
                    raise ValueError(f'New dim header length must be {self.shape[dim]}')

            for key in header:
                if isinstance(header[key], list):
                    size_chk(header[key])
                elif isinstance(header[key], dict)\
                        and 'Value' in header[key]:
                    size_chk(header[key]['Value'])

        current_hdr_ext = self.hdr_ext
        current_hdr_ext.set_dim_info(dim - 4, tag, info=info, hdr=header)
        self.hdr_ext = current_hdr_ext

    def copy(self, remove_dim=None):
        """Return a copy of this image, optionally with a dimension removed.

        :param remove_dim: dimension index (4, 5, 6) or tag. None iterates over all indices., defaults to None
        :type remove_dim: str or int, optional
        :return: Copy of object
        :rtype: NIFTI_MRS
        """
        if remove_dim:
            dim = self._dim_tag_to_index(remove_dim)
            reduced_data = self[:].take(0, axis=dim)
            new_obj = NIFTI_MRS(reduced_data, header=self.header)
            new_obj._filename = self.filename

            # Modify the dim information in
            hdr_ext = self.hdr_ext
            hdr_ext.remove_dim_info(dim - 4)
            new_obj.hdr_ext = hdr_ext.to_dict()

            return new_obj
        else:
            return NIFTI_MRS(self[:], header=self.header)

    def save(self, filepath):
        """Save NIfTI-MRS to file

        :param filepath: Name and path of save loaction
        :type filepath: str or pathlib.Path
        """
        # Ensure final copy of header extension is loaded into Image object
        self._save_hdr_ext()

        # Run validation
        validator.validate_nifti_mrs(self._image)

        # Save underlying image object to file
        self._image.save(filepath)

    # Methods for iteration over dimensions
    def iterate_over_dims(self, dim=None, iterate_over_space=False, reduce_dim_index=False, voxel_index=None):
        """Return generator to iterate over all indices or one dimension (and FID).

        :param dim: None, dimension index (4, 5, 6) or tag. None iterates over all indices. Defaults to None
        :type dim: str or int, optional
        :param iterate_over_space: If True also iterate over spatial dimension, defaults to False
        :type iterate_over_space: bool, optional
        :param reduce_dim_index: If True the returned slice index will have the selected dimension removed.
            Defaults to False.
        :type reduce_dim_index: bool, optional
        :param voxel_index: slice or tuple of first three spatial dimensions., defaults to None
        :type voxel_index: slice or tuple, optional
        :return: yeildsarray of sliced data
        :rtype: np.array
        :return: data location slice object.
        :rtype: slice
        """

        data = self[:]
        dim = self._dim_tag_to_index(dim)

        # Convert indicies to slices to preserve singleton dimensions
        if voxel_index is not None:
            tmp = []
            for vi in voxel_index:
                if isinstance(vi, slice):
                    tmp.append(vi)
                elif isinstance(vi, int):
                    tmp.append(slice(vi, vi + 1))
                else:
                    raise TypeError('voxel index elements must be slice or int type.')
            voxel_index = tuple(tmp)

        def calc_slice_idx(idx):
            if iterate_over_space:
                slice_obj = list(idx[:3]) + [slice(None), ] + list(idx[3:])
            else:
                slice_obj = [slice(None), slice(None), slice(None), slice(None)]\
                    + list(idx[0:])
            if dim is not None and not reduce_dim_index:
                slice_obj.insert(dim + 1, slice(None))
            return tuple(slice_obj)

        if isinstance(dim, (int, str)):
            # Move FID dim to last
            data = np.moveaxis(data, 3, -1)
            dim -= 1
            # Move identified dim to last
            data = np.moveaxis(data, dim, -1)

            if voxel_index is not None:
                voxel_index
                data = data[voxel_index]

            if iterate_over_space:
                iteration_skip = -2
            else:
                data = np.moveaxis(data, (0, 1, 2), (-5, -4, -3))
                iteration_skip = -5

            for idx in np.ndindex(data.shape[:iteration_skip]):
                yield data[idx], calc_slice_idx(idx)

        elif dim is None:
            # Move FID dim to last
            data = np.moveaxis(data, 3, -1)

            if voxel_index is not None:
                data = data[voxel_index]

            if iterate_over_space:
                iteration_skip = -1
            else:
                data = np.moveaxis(data, (0, 1, 2), (-4, -3, -2))
                iteration_skip = -4

            for idx in np.ndindex(data.shape[:iteration_skip]):
                yield data[idx], calc_slice_idx(idx)

        else:
            raise TypeError('dim should be int or a string matching one of the dim tags.')

    def iterate_over_spatial(self):
        """Iterate over spatial voxels yeilding a data array the shape of the FID and any higher dimensions + index.

        :yield: Complex FID data with any higher dimensions. Index to data.
        :rtype: tuple
        """
        data = self[:]

        def calc_slice_idx(idx):
            slice_obj = list(idx[:3]) + [slice(None), ] * (data.ndim - 3)
            return tuple(slice_obj)

        for idx in np.ndindex(data.shape[:3]):
            yield self[idx], calc_slice_idx(idx)

    def dynamic_hdr_vals(self):
        """Return representations of the dynamic header values

        :return: List of dicts containing labeled header parameters
        :return: List of tuples containing header values
        :return: Flattened numpy array for each generated spectrum containing header values
        """
        def list_of_dict_from_dim(dim_hdr, size):
            """Form a list of dicts where each index of the list corresponds to one element"""
            out = []
            for idx in range(size):
                tmp_dict = {}
                for key in dim_hdr:
                    # Handle the short form!
                    if 'increment' in dim_hdr[key]:
                        start = dim_hdr[key]['start']
                        inc = dim_hdr[key]['increment']
                        tmp_dict.update({key: start + inc * idx})
                    else:
                        tmp_dict.update({key: dim_hdr[key][idx]})
                out.append(tmp_dict)
            return out

        def sort_output(hdr_list):
            if len(hdr_list) == 1:
                X = np.meshgrid(hdr_list[0])
                tvar = [x for x in X[0]]
            elif len(hdr_list) == 2:
                X, Y = np.meshgrid(hdr_list[0], hdr_list[1], indexing='ij')
                tvar = [dict(x, **y) for x, y in zip(X.ravel(), Y.ravel())]
            elif len(hdr_list) == 3:
                X, Y, Z = np.meshgrid(hdr_list[0], hdr_list[1], hdr_list[2], indexing='ij')
                tvar = [dict(x, **y, **z) for x, y, z in zip(X.ravel(), Y.ravel(), Z.ravel())]
            return tvar

        def convert_to_tuples(dict_list):
            out_list = []
            for dl in dict_list:
                tl = []
                for key in dl:
                    tl.append(dl[key])
                out_list.append(tuple(tl))
            return out_list

        all_dim_hdrs_dict = []
        for dim in range(5, 8):
            if f'dim_{dim}_header' in self.hdr_ext:
                all_dim_hdrs_dict.append(
                    list_of_dict_from_dim(self.hdr_ext[f'dim_{dim}_header'], self.shape[dim - 1]))

        tvar_dict = sort_output(all_dim_hdrs_dict)
        tvar_tuple = convert_to_tuples(tvar_dict)

        tvar_dict2 = np.asarray(tvar_dict, dtype=object).reshape(self.shape[4:])
        tvar_tuple2 = np.empty_like(tvar_dict2).flatten()
        for idx, elm in enumerate(tvar_tuple):
            tvar_tuple2[idx] = elm
        tvar_array = np.asarray(tvar_tuple, dtype=object).reshape(np.prod(self.shape[4:]), len(tvar_tuple[0]))

        return tvar_dict2, tvar_tuple2.reshape(self.shape[4:]), tvar_array



"""Tools for reshaping the higher dimensions of NIfTI-MRS

    Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
    Copyright (C) 2021 University of Oxford
"""
from nifti_mrs.nifti_mrs import NIFTI_MRS

import numpy as np

# TO DO:
# def _reshape_hdr(dynamic_hdr, target):

#     return reshaped_hdr


def reshape(nmrs, reshape, d5=None, d6=None, d7=None):
    """Reshape the higher dimensions (5-7) of an nifti-mrs file.
    Uses numpy reshape syntax to reshape. Use -1 for automatic sizing.

    If the dimension exists after reshaping a tag is required. If None is passed
    but one already exists no change will be made. If no value exists then an
    exception will be raised.

    :param nmrs: Input NIfTI-MRS file
    :type nmrs: NIFTI_MRS
    :param reshape: Tuple of target sizes in style of numpy.reshape, higher dimensions only.
    :type reshape: tuple
    :param d5: Dimension tag to set dim_5, defaults to None
    :type d5: str, optional
    :param d6: Dimension tag to set dim_6, defaults to None
    :type d6: str, optional
    :param d7: Dimension tag to set dim_7, defaults to None
    :type d7: str, optional
    """

    shape = nmrs[:].shape[0:4]
    shape += reshape
    nmrs_reshaped = NIFTI_MRS(np.reshape(nmrs[:], shape), header=nmrs.header)

    # reshpaed_hrd = _reshape_hdr(nmrs_reshaped.dynamic_hdr_vals[2],)

    # Note numerical index is N-1
    if d5:
        nmrs_reshaped.set_dim_tag(4, d5)
    elif nmrs_reshaped.ndim > 4\
            and nmrs.dim_tags[0] is None:
        raise TypeError(f'An appropriate d5 dim tag must be given as ndim = {nmrs_reshaped.ndim}.')
    if d6:
        nmrs_reshaped.set_dim_tag(5, d6)
    elif nmrs_reshaped.ndim > 5\
            and nmrs.dim_tags[1] is None:
        raise TypeError(f'An appropriate d6 dim tag must be given as ndim = {nmrs_reshaped.ndim}.')
    if d7:
        nmrs_reshaped.set_dim_tag(6, d7)
    elif nmrs_reshaped.ndim > 6\
            and nmrs.dim_tags[2] is None:
        raise TypeError(f'An appropriate d7 dim tag must be given as ndim = {nmrs_reshaped.ndim}.')

    return nmrs_reshaped



"""Generate new NIfTI-MRS objects and files from data blocks.

Copyright William Clarke, University of Oxford, 2023
"""
import nibabel as nib
import numpy as np

from nifti_mrs.definitions import nifti_mrs_version
from nifti_mrs.nifti_mrs import NIFTI_MRS
from nifti_mrs.hdr_ext import Hdr_Ext


def _checkCFUnits(cf, units='Hz'):
    """ Check the units of central frequency and adjust if required."""
    # Assume cf in Hz > 1E5, if it isn't assume that user has passed in MHz
    if cf < 1E5:
        if units.lower() == 'hz':
            cf *= 1E6
        elif units.lower() == 'mhz':
            pass
        else:
            raise ValueError('Only Hz or MHz defined')
    else:
        if units.lower() == 'hz':
            pass
        elif units.lower() == 'mhz':
            cf /= 1E6
        else:
            raise ValueError('Only Hz or MHz defined')
    return cf


def _default_affine():
    """Return NIfTI-MRS default affine matrix"""
    default = np.eye(4) * 10000.0
    default[3, 3] = 1.0
    return default


def gen_nifti_mrs(data, dwelltime, spec_freq, nucleus='1H', affine=None, dim_tags=[None, None, None], nifti_version=2):
    """Generate NIfTI-MRS from data and required metadata

    :param data: Complex-typed numpy array of at least 4 dimensions (max 7)
    :type data: numpy.array
    :param dwelltime: Spectral (4th dimension) dwelltime in seconds
    :type dwelltime: float
    :param spec_freq: Spectrometer Frequency in MHz
    :type spec_freq: float
    :param nucleus: Resonant Nucleus string (e.g. 1H, 31P, 2H), defaults to '1H'
    :type nucleus: str, optional
    :param affine: 4x4 orientation/position affine, defaults to None which will use default (scaled identity).
    :type affine: numpy.array, optional
    :param dim_tags: List of dimension tags (e.g. DIM_DYN), defaults to [None, None, None]
    :type dim_tags: list, optional
    :param nifti_version: Version of NIfTI header format, defaults to 2
    :type nifti_version: int, optional
    :return: NIfTI-MRS object
    :rtype: nifti_mrs.nifti_mrs.NIFTI_MRS
    """
    # Create header_ext
    hdr_ext = Hdr_Ext(
        _checkCFUnits(spec_freq, units='MHz'),
        nucleus,
        dimensions=data.ndim)

    for idx, tag in enumerate(dim_tags):
        if tag is not None\
                and idx + 4 < data.ndim:
            hdr_ext.set_dim_info(idx, tag)

    # Pass data and hdr_ext to gen_nifti_mrs_hdr_ext
    return gen_nifti_mrs_hdr_ext(
        data,
        dwelltime,
        hdr_ext,
        affine=affine,
        nifti_version=nifti_version)


def gen_nifti_mrs_hdr_ext(data, dwelltime, hdr_ext, affine=None, nifti_version=2):
    """Generate NIfTI-MRS from data and header extension object

    :param data: Complex-typed numpy array of at least 4 dimensions (max 7)
    :type data: numpy.array
    :param dwelltime: Spectral (4th dimension) dwelltime in seconds
    :type dwelltime: float
    :param hdr_ext: Populated NIfTI-MRS header extension
    :type hdr_ext: nifti_mrs.hdr_ext.Hdr_Ext
    :param affine: 4x4 orientation/position affine, defaults to None which will use default (scaled identity).
    :type affine: numpy.array, optional
    :param dim_tags: List of dimension tags (e.g. DIM_DYN), defaults to [None, None, None]
    :type dim_tags: list, optional
    :param nifti_version: Version of NIfTI header format, defaults to 2
    :type nifti_version: int, optional
    :return: NIfTI-MRS object
    :rtype: nifti_mrs.nifti_mrs.NIFTI_MRS
    """

    if not np.iscomplexobj(data):
        raise ValueError('data must be complex')
    if data.ndim < 4 or data.ndim > 7:
        raise ValueError(f'data must have between 4 and 7 dimensions, currently has {data.ndim}')

    if affine is None:
        affine = _default_affine()

    # Create a nifti image as a short cut to creating a header
    if nifti_version == 1:
        tmp_img = nib.nifti1.Nifti1Image(
            data,
            affine=affine)
        header = tmp_img.header
    elif nifti_version == 2:
        tmp_img = nib.nifti2.Nifti2Image(
            data,
            affine=affine)
        header = tmp_img.header
    else:
        raise ValueError('nifti_version must be 1 or 2')

    # Orientation/position info
    header.set_qform(affine)
    header.set_sform(affine)

    # Touch up header with required NIfTI-MRS metadata
    header['pixdim'][4] = dwelltime
    v_major = nifti_mrs_version[0]
    v_minor = nifti_mrs_version[1]
    header['intent_name'] = f'mrs_v{v_major}_{v_minor}'.encode()

    # Add header extension to header
    json_s = hdr_ext.to_json()
    extension = nib.nifti1.Nifti1Extension(44, json_s.encode('UTF-8'))
    header.extensions.append(extension)

    return NIFTI_MRS(data, header=header)



""" mrs_tools - top level script for calling NIfTI-MRS handling tools

Author: William Clarke <william.clarke@ndcn.ox.ac.uk>

Copyright (C) William Clarke, 2023, University of Oxford
"""

import argparse
from pathlib import Path

from nifti_mrs import __version__


def main(import_args=None):
    # Parse command-line arguments
    p = argparse.ArgumentParser(description="NIfTI-MRS (Magnetic Resonance Spectroscopy) tools")

    p.add_argument('-v', '--version', action='version', version=__version__)

    sp = p.add_subparsers(title='subcommands',
                          description='Available tools',
                          required=True,
                          dest='subcommand')

    # Info tool
    infoparser = sp.add_parser(
        'info',
        help='Information about the NIfTI-MRS file.')
    infoparser.add_argument(
        'file',
        type=Path,
        metavar='FILE or list of FILEs',
        help='NIfTI MRS file(s)', nargs='+')
    infoparser.set_defaults(func=info)

    # Vis tool
    visparser = sp.add_parser(
        'vis',
        help='Quick visualisation of a NIfTI-MRS file or FSL-MRS basis set.')
    visparser.add_argument('file', type=Path, metavar='FILE or DIR',
                           help='NIfTI file or directory of basis sets')
    visparser.add_argument('--ppmlim', default=(.2, 4.2), type=float,
                           nargs=2, metavar=('LOW', 'HIGH'),
                           help='limit the fit to a freq range (default=(.2,4.2))')
    visparser.add_argument('--mask', default=None, type=str, help='Mask for MRSI')
    visparser.add_argument('--save', default=None, type=str, help='Save fig to path')
    visparser.add_argument('--display_dim', default=None, type=str,
                           help='NIFTI-MRS tag. Do not average across this dimension.')
    visparser.add_argument('--no_mean', action="store_false",
                           help='Do not plot the mean signal line in the case of multiple spectra.')
    visparser.set_defaults(func=vis)

    # Merge tool - Merge NIfTI MRS along higher dimensions
    mergeparser = sp.add_parser(
        'merge',
        help='Merge NIfTI-MRS along higher dimensions.')
    merge_req = mergeparser.add_argument_group('required arguments')
    merge_req.add_argument('--files', type=Path, nargs='+', required=True,
                           help='List of files to merge')
    merge_req.add_argument('--dim', type=str, required=True,
                           help='NIFTI-MRS dimension tag to merge across.')
    mergeparser.add_argument('--output', type=Path, default=Path('.'),
                             help='output folder (defaults to current directory)')
    mergeparser.add_argument('--filename', type=str,
                             help='Override output file name.')
    mergeparser.set_defaults(func=merge)

    # Split tool
    splitparser = sp.add_parser(
        'split',
        help='Split NIfTI-MRS along higher dimensions.')
    split_req = splitparser.add_argument_group('required arguments')
    split_req.add_argument('--file', type=Path, required=True,
                           help='File to split')
    split_req.add_argument('--dim', type=str, required=True,
                           help='NIFTI-MRS dimension tag to split across.')
    group = splitparser.add_mutually_exclusive_group(required=True)
    group.add_argument('--indices', type=int, nargs='+',
                       help='List of indices to extract into second file.'
                            ' All indices are zero-indexed.')
    group.add_argument('--index', type=int,
                       help='Index to split at (split after index, zero-indexed).')
    splitparser.add_argument('--output',
                             required=False, type=Path, default=Path('.'),
                             help='output folder (defaults to current directory)')
    splitparser.add_argument('--filename', type=str,
                             help='Override output file names.')
    splitparser.set_defaults(func=split)

    # Reorder tool
    reorderparser = sp.add_parser(
        'reorder',
        help='Reorder higher dimensions of NIfTI-MRS.')
    reord_req = reorderparser.add_argument_group('required arguments')
    reord_req.add_argument('--file', type=Path, required=True,
                           help='File to reorder')
    reord_req.add_argument('--dim_order', type=str, nargs='+', required=True,
                           help='NIFTI-MRS dimension tags in desired order. '
                                'Enter as strings (min:1, max:3). '
                                'Can create singleton dimension at end.')
    reorderparser.add_argument('--output',
                               required=False, type=Path, default=Path('.'),
                               help='output folder (defaults to current directory)')
    reorderparser.add_argument('--filename', type=str,
                               help='Override output file names.')
    reorderparser.set_defaults(func=reorder)

    # conjugate tool
    conjparser = sp.add_parser(
        'conjugate',
        help='Conjugate data to correct phase/frequency convention in a NIfTI-MRS file.')
    conj_req = conjparser.add_argument_group('required arguments')
    conj_req.add_argument('--file', type=Path, required=True,
                          help='File to conjugate')
    conjparser.add_argument('--output',
                            required=False, type=Path, default=Path('.'),
                            help='output folder (defaults to current directory)')
    conjparser.add_argument('--filename', type=str,
                            help='Override output file names.')
    conjparser.set_defaults(func=conj)

    # Parse command-line arguments
    args = p.parse_args(import_args)

    # Call function
    args.func(args)


def info(args):
    """Prints basic information about NIfTI-MRS files
    :param args: Argparse interpreted arguments
    :type args: Namespace
    """
    from nifti_mrs.nifti_mrs import NIFTI_MRS
    from mrs_tools.constants import GYRO_MAG_RATIO

    for file in args.file:
        data = NIFTI_MRS(file)

        print(f'\nRead file {file.name} ({file.parent.resolve()}).')
        print(f'NIfTI-MRS version {data.nifti_mrs_version}')
        print(f'Data shape {data.shape}')
        print(f'Dimension tags: {data.dim_tags}')

        print(f'Spectrometer Frequency: {data.spectrometer_frequency[0]} MHz')
        print(f'Dwelltime (Bandwidth): {data.dwelltime:0.3E}s ({data.bandwidth:0.0f} Hz)')
        print(f'Nucleus: {data.nucleus[0]}')
        if data.nucleus[0] in GYRO_MAG_RATIO:
            field_strength = data.spectrometer_frequency[0] / GYRO_MAG_RATIO[data.nucleus[0]]
            print(f'Field Strength: {field_strength:0.2f} T')
        print()


def vis(args):
    """Visualiser for NIfTI-MRS files

    Visualisation requires FSL-MRS to be installed.

    TO DO: Make fall-back visualisation.

    :param args: Argparse interpreted arguments
    :type args: Namespace
    """
    try:
        from fsl_mrs.utils.plotting import plot_spectrum, plot_spectra
        from fsl_mrs.utils.mrs_io import read_FID, read_basis
        from fsl_mrs.utils.mrs_io.main import FileNotRecognisedError
        import matplotlib.pyplot as plt
        from fsl_mrs.utils.preproc import nifti_mrs_proc
    except ImportError:
        raise ImportError(
            "mrs_tools vis requires FSL-MRS tools to be installed. "
            "See fsl-mrs.com for installation instructions.")

    import numpy as np
    import nibabel as nib

    # Single nifti file
    def vis_nifti_mrs(file):
        data = read_FID(file)
        if data.ndim > 4 \
                and 'DIM_COIL' in data.dim_tags\
                and args.display_dim != 'DIM_COIL':
            print('Performing coil combination')
            data = nifti_mrs_proc.coilcombine(data)

        if np.prod(data.shape[:3]) == 1:
            # SVS
            if args.display_dim:
                for idx in range(data.ndim - 4):
                    if data.dim_tags[idx] != args.display_dim:
                        print(f'Averaging {data.dim_tags[idx]}')
                        data = nifti_mrs_proc.average(data, data.dim_tags[idx])

                fig = plot_spectra(data.mrs(), ppmlim=args.ppmlim, plot_avg=args.no_mean)
            else:
                while data.ndim > 4:
                    print(f'Averaging {data.dim_tags[0]}')
                    data = nifti_mrs_proc.average(data, data.dim_tags[0])
                fig = plot_spectrum(data.mrs(), ppmlim=args.ppmlim)
            if args.save is not None:
                fig.savefig(args.save)
            else:
                plt.show()

        else:
            while data.ndim > 4:
                print(f'Averaging {data.dim_tags[0]}')
                data = nifti_mrs_proc.average(data, data.dim_tags[0])

            mrsi = data.mrs()
            if args.mask is not None:
                mask_hdr = nib.load(args.mask)
                mask = np.asanyarray(mask_hdr.dataobj)
                if mask.ndim == 2:
                    mask = np.expand_dims(mask, 2)
                mrsi.set_mask(mask)
            mrsi.plot()

    # Some logic to figure out what we are dealing with
    p = args.file
    nifti_files = list(p.glob('*.nii*'))

    # Identify BASIS
    if (p.is_dir() and len(nifti_files) == 0)\
            or p.suffix.upper() == '.BASIS':

        # Some heuristics
        if p.is_dir():
            conj = True
        else:
            conj = False

        basis = read_basis(args.file)
        _ = basis.plot(ppmlim=args.ppmlim, conjugate=conj)
        if args.save is not None:
            plt.savefig(args.save)
        else:
            plt.show()

    # Identify directory of nifti files
    elif p.is_dir() and len(nifti_files) > 0:
        raise ValueError('mrs_tools vis should be called on a single'
                         ' NIFTI-MRS file, not a directory (unless'
                         ' it contains basis files).')

    elif p.is_file():
        vis_nifti_mrs(p)

    else:
        try:
            vis_nifti_mrs(p)
        except FileNotRecognisedError as exc:
            raise FileNotFoundError(
                f"No file or directory '{p}' found."
                " Please specify correct file extension (e.g. nii.gz) if there is one.")\
                from exc


def merge(args):
    """Merges one or more NIfTI-MRS files along a specified dimension
    :param args: Argparse interpreted arguments
    :type args: Namespace
    """
    import nifti_mrs.tools as nmrs_tools
    from nifti_mrs.nifti_mrs import NIFTI_MRS
    # 1. Load the files
    if len(args.files) < 2:
        raise ValueError('Files argument must provide two or more files to merge.')

    to_concat = []
    concat_names = []
    for fp in args.files:
        concat_names.append(fp.with_suffix('').with_suffix('').name)
        to_concat.append(NIFTI_MRS(fp))

    # 2. Merge the files
    merged = nmrs_tools.merge(to_concat, args.dim)

    # 3. Save the output file
    if args.filename:
        file_out = args.output / args.filename
    else:
        file_out = args.output / ('_'.join(concat_names) + '_merged')
    merged.save(file_out)


def split(args):
    """Splits a NIfTI-MRS file into two along a specified dimension
    :param args: Argparse interpreted arguments
    :type args: Namespace
    """
    import nifti_mrs.tools as nmrs_tools
    from nifti_mrs.nifti_mrs import NIFTI_MRS
    # 1. Load the file
    to_split = NIFTI_MRS(args.file)
    split_name = args.file.with_suffix('').with_suffix('').name

    # 2. Merge the files
    if args.index is not None:
        split_1, split_2 = nmrs_tools.split(to_split, args.dim, args.index)
    elif args.indices:
        split_1, split_2 = nmrs_tools.split(to_split, args.dim, args.indices)

    # 3. Save the output file
    if args.index is not None:
        first_name = '_low'
        second_name = '_high'
    elif args.indices:
        first_name = '_others'
        second_name = '_selected'

    if args.filename:
        file_out_1 = args.output / (args.filename + first_name)
        file_out_2 = args.output / (args.filename + second_name)
    else:
        file_out_1 = args.output / (split_name + first_name)
        file_out_2 = args.output / (split_name + second_name)
    split_1.save(file_out_1)
    split_2.save(file_out_2)


def reorder(args):
    """Reorders the higher dimensions of a NIfTI-MRS file
    :param args: Argparse interpreted arguments
    :type args: Namespace
    """
    import nifti_mrs.tools as nmrs_tools
    from nifti_mrs.nifti_mrs import NIFTI_MRS
    # 1. Load the file
    to_reorder = NIFTI_MRS(args.file)
    reorder_name = args.file.with_suffix('').with_suffix('').name

    # 2. Merge the files
    dim_order = args.dim_order
    while len(dim_order) < 3:
        dim_order.append(None)
    reordered = nmrs_tools.reorder(to_reorder, args.dim_order)

    # 3. Save the output file
    if args.filename:
        file_out = args.output / args.filename
    else:
        file_out = args.output / (reorder_name + '_reordered')
    reordered.save(file_out)


def conj(args):
    """Conjugate the data in a nifti-mrs file

    :param args: Argparse interpreted arguments
    :type args: Namespace
    """
    import nifti_mrs.tools as nmrs_tools
    from nifti_mrs.nifti_mrs import NIFTI_MRS
    # 1. Load the file
    infile = NIFTI_MRS(args.file)
    name = args.file.with_suffix('').with_suffix('').name

    # 2. conjugate the file
    outfile = nmrs_tools.conjugate(infile)

    # 3. Save the output file
    if args.filename:
        file_out = args.output / args.filename
    else:
        file_out = args.output / name
    outfile.save(file_out)



from .definitions import dimension_tags, standard_defined
import json


class Hdr_Ext:
    """Class to hold meta data stored in a NIfTI MRS header extension.
    Required fields must be passed to initialise,
    Default dimension information automatically generated, but may be modified by set_dim_info method.
    Standard defined meta-data and user-defined data can be added using set_standard_def and
    set_user_def respectively.
    """
    def __init__(self, spec_frequency, resonant_nucleus, dimensions=7):
        """Initialise NIfTI-MRS header extension object with the two mandatory bits of meta-data.

        :param spec_frequency: Spectrometer frequency in MHz
        :type spec_frequency: float
        :param resonant_nucleus: Resonant nucleus e.g. '1H'
        :type resonant_nucleus: str
        :param dimensions: Number of dimensions in image. Controls number of dim tags in extension. Defaults to 7
        :type dimensions: int, optional
        """
        if isinstance(spec_frequency, float):
            self.SpectrometerFrequency = [spec_frequency, ]
        elif isinstance(spec_frequency, (list, tuple))\
                and isinstance(spec_frequency[0], float):
            self.SpectrometerFrequency = spec_frequency
        else:
            raise ValueError('spec_frequency must be a float or array of floats.')

        if isinstance(resonant_nucleus, str):
            self.ResonantNucleus = [resonant_nucleus, ]
        elif isinstance(resonant_nucleus, (list, tuple))\
                and isinstance(resonant_nucleus[0], str):
            self.ResonantNucleus = resonant_nucleus
        else:
            raise ValueError('resonant_nucleus must be a string or array of strings.')

        self._dim_info = [{"tag": "DIM_COIL", "info": None, "hdr": None},
                          {"tag": "DIM_DYN", "info": None, "hdr": None},
                          {"tag": "DIM_INDIRECT_0", "info": None, "hdr": None}]

        self._standard_data = {}
        self._user_data = {}
        self.dimensions = dimensions

    @classmethod
    def from_header_ext(cls, hdr_ext_dict, dimensions=7):

        optional_dict = dict(hdr_ext_dict)
        obj = cls(
            hdr_ext_dict['SpectrometerFrequency'],
            hdr_ext_dict['ResonantNucleus'])
        optional_dict.pop('SpectrometerFrequency')
        optional_dict.pop('ResonantNucleus')

        for idx in range(5, 8):
            if f'dim_{idx}' in hdr_ext_dict:
                tag = hdr_ext_dict[f'dim_{idx}']
                optional_dict.pop(f'dim_{idx}')
                if f'dim_{idx}_info' in hdr_ext_dict:
                    info = hdr_ext_dict[f'dim_{idx}_info']
                    optional_dict.pop(f'dim_{idx}_info')
                else:
                    info = None
                if f'dim_{idx}_header' in hdr_ext_dict:
                    hdr = hdr_ext_dict[f'dim_{idx}_header']
                    optional_dict.pop(f'dim_{idx}_header')
                else:
                    hdr = None
                obj.set_dim_info(idx - 5, tag, info=info, hdr=hdr)

        for key in optional_dict:
            if key in standard_defined:
                obj.set_standard_def(key, hdr_ext_dict[key])
            else:
                obj.set_user_def(key=key, value=hdr_ext_dict[key])

        obj.dimensions = dimensions

        return obj

    def set_dim_info(self, dim, tag, info=None, hdr=None):
        """Set information associated with the optional, higher data dimensions.

        :param dim:  May be (0,1,2) or ("5th","6th","7th")
        :type dim: str or int
        :param tag: Must be one of the defined dimension tag strings. E.g. DIM_DYN
        :type tag: str
        :param info: Optional, free-form for documentation, defaults to None
        :type info: str, optional
        :param hdr: Dict containing relevant header value names and values. Defaults to None
        :type hdr: dict, optional
        """
        if tag not in dimension_tags:
            raise ValueError("tag must be one of the defined dimension tag.")

        new_info = {"tag": tag,
                    "info": info,
                    "hdr": hdr}

        if dim in (0, 1, 2):
            self._dim_info[dim] = new_info
        elif dim in ("5th", "6th", "7th"):
            if dim == "5th":
                self._dim_info[0] = new_info
            elif dim == "6th":
                self._dim_info[1] = new_info
            elif dim == "7th":
                self._dim_info[2] = new_info
        else:
            raise ValueError('dim must be 0,1,2 or "5th","6th","7th".')

    def remove_dim_info(self, dim):
        """Set a dimension's information to None

        :param dim: 0,1,2 or "5th","6th","7th"
        :type dim: str or int
        """
        if dim in (0, 1, 2):
            pass
        elif dim in ("5th", "6th", "7th"):
            if dim == "5th":
                dim = 0
            elif dim == "6th":
                dim = 1
            elif dim == "7th":
                dim = 2
        else:
            raise ValueError('dim must be 0,1,2 or "5th","6th","7th".')

        self._dim_info.pop(dim)
        self._dim_info.append({"tag": None, "info": None, "hdr": None})

    def set_standard_def(self, key, value):
        """Add a single standard-defined bit of meta-data to the object."""
        if key not in standard_defined:
            raise ValueError("key must be one of the standard-defined keys.")

        self._standard_data[key] = value

    def set_user_def(self, all_keys=None, key=None, value=None, doc=None):
        """Add user defined meta data keys to the header extension.
        Pass dict as kwarg all_keys to set all key/value pairs, or
        add keys and values one at a time using key, value and doc.
        """

        if all_keys is not None:
            self._user_data = all_keys
        else:
            if key in standard_defined:
                raise ValueError("key must not be one of the standard-defined keys.")

            if isinstance(value, dict):
                self._user_data[key] = value
                self._user_data[key].update({'Description': doc})
            else:
                self._user_data[key] = {
                    'Value': value,
                    'Description': doc}

    def remove_standard_def(self, key):
        """Remove key from list of standard defined key-value pairs

        :param key: Key name
        :type key: str
        """
        if key not in self.current_keys:
            raise KeyError(f'{key} is not defined in the header extension.')
        self._standard_data.pop(key)

    def remove_user_def(self, key):
        """Remove key from list of user defined key-value pairs

        :param key: Key name
        :type key: str
        """
        if key not in self.current_keys:
            raise KeyError(f'{key} is not defined in the header extension.')
        self._user_data.pop(key)

    def to_dict(self):
        """Generate dictionay representation from properties."""

        # Required meta-data
        out_dict = {'SpectrometerFrequency': self.SpectrometerFrequency,
                    'ResonantNucleus': self.ResonantNucleus}

        # Dimension information
        if self.dimensions < 4:
            raise ValueError('dimensions must be 4 or greater')
        elif self.dimensions == 4:
            pass
        else:
            update_dict = {}
            for idx in range(5, self.dimensions + 1):
                update_dict[f'dim_{idx}'] = self._dim_info[idx - 5]['tag']
                if self._dim_info[idx - 5]['info'] is not None:
                    update_dict[f'dim_{idx}_info'] = self._dim_info[idx - 5]['info']
                if self._dim_info[idx - 5]['hdr'] is not None:
                    update_dict[f'dim_{idx}_header'] = self._dim_info[idx - 5]['hdr']

            out_dict.update(update_dict)

        # Add standard defined
        out_dict.update(self._standard_data)

        # Add user defined
        out_dict.update(self._user_data)

        return out_dict

    @property
    def current_keys(self):
        return self.to_dict().keys()

    def to_json(self):
        return json.dumps(self.to_dict())

    # For dict-like behaviour
    def __getitem__(self, key):
        return self.to_dict()[key]

    def __contains__(self, key):
        return key in self.to_dict()

    def __str__(self) -> str:
        return self.to_json()

    def __repr__(self) -> str:
        return str(self)

    def __iter__(self):
        yield from self.current_keys

    def copy(self):
        from copy import deepcopy
        return deepcopy(self)

    def __eq__(self, other):
        if isinstance(other, Hdr_Ext):
            return self.to_dict() == other.to_dict()
        elif isinstance(other, dict):
            return self.to_dict() == other
        else:
            raise NotImplementedError('Equality can only be tested with dict or Hdr_Ext object.')



""" Miscellaneous tools for NIfTI-MRS

    Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
    Copyright (C) 2021 University of Oxford
"""

import numpy as np

from nifti_mrs.nifti_mrs import NIFTI_MRS


def conjugate(nmrs):
    """Conjugate a nifti-mrs object.

    :param nmrs: NIFTI_MRS object to conjugate
    :type nmrs: NIFTI_MRS
    :return: Conjugated NIFTI_MRS
    :rtype: NIFTI_MRS
    """

    return NIFTI_MRS(np.conjugate(nmrs[:]), header=nmrs.header)



import json
from .definitions import dimension_tags, standard_defined
import numpy as np
import re


class Error(Exception):
    """Base class for other exceptions"""
    pass


class headerExtensionError(Error):
    """Raised if problems with header extension are found."""
    pass


class niftiHeaderError(Error):
    """Raised if problems with nifti header are found."""
    pass


class niftiDataError(Error):
    """Raised if problems with nifti data are found."""
    pass


def validate_nifti_mrs(nifti_mrs):
    """Validate a full NIfTI MRS image."""

    # Validate data
    validate_nifti_data(nifti_mrs[:])

    # Validate nifit header
    validate_nifti_header(nifti_mrs.header)

    # Validate header extension
    validate_hdr_ext(nifti_mrs.header.extensions[0].get_content(),
                     nifti_mrs.ndim)


def validate_nifti_data(nifti_img_data):
    """Validate the data inside a nibabel nifti image
    1. Check data is complex
    2. Check number of dimensions is at least 4 but less than 8.
    """

    # 1. Check for complexity
    if not np.iscomplexobj(nifti_img_data):
        raise niftiDataError('Data must be complex.')

    # 2. Check for between 4 and 7 dimensions
    if nifti_img_data.ndim < 4\
            or nifti_img_data.ndim > 7:
        raise niftiDataError('Data must have between 4 and 7 dimensions.'
                             f' It has {nifti_img_data.ndim}.')


def validate_nifti_header(nifti_header):
    """Validate the header of a nibabel nifti image
    Check data type is complex
    Check orientation data. - How? TO DO
    Check dwell time
    Check intent name
    """
    if nifti_header.get_data_dtype() not in (complex, np.complex64, np.complex128):
        raise niftiHeaderError('Data type is not complex.')

    if nifti_header['pixdim'][4] <= 0 or nifti_header['pixdim'][4] > 1.0:
        raise niftiHeaderError(f'Dwell time ({nifti_header["pixdim"][4]}) is unrealistic')

    intent_ptrn = re.compile(r'mrs_v\d+_\d+')
    intent_str = nifti_header.get_intent()[2]
    if intent_ptrn.match(intent_str) is None:
        raise niftiHeaderError(f'Intent string ({intent_str}) does not match "mrs_vMajor_minor".')


def validate_hdr_ext(header_ex, data_dimensions):
    """ Validate the header extension
    1. Check that it is json formatted string.
    2. Check that it contains the required meta-data
    3. Check that it contains any required dimension information.
    4. Check that standard-defined data is of correct type.
    Inputs:

    """
    # 1. Check that header_ext is json
    try:
        json_dict = json.loads(header_ex)
    except json.JSONDecodeError as exc:
        raise headerExtensionError("Header extension is not json deserialisable.") from exc

    # 2. Check the two required bits of meta-data
    if "SpectrometerFrequency" in json_dict:
        if not isinstance(json_dict["SpectrometerFrequency"], (list, tuple))\
                and not isinstance(json_dict["SpectrometerFrequency"][0], float):

            raise headerExtensionError("SpectrometerFrequency must be list of floats.")
    else:
        raise headerExtensionError("Header extension must contain SpectrometerFrequency.")

    if "ResonantNucleus" in json_dict:
        if not isinstance(json_dict["ResonantNucleus"], (list, tuple))\
                and not isinstance(json_dict["ResonantNucleus"][0], str):

            raise headerExtensionError("ResonantNucleus must be list of strings.")
    else:
        raise headerExtensionError("Header extension must contain ResonantNucleus.")

    # 3. Dimension information
    for ddx in range(5, 8):
        if data_dimensions > (ddx - 1):
            if f"dim_{ddx}" in json_dict:
                if json_dict[f"dim_{ddx}"] not in dimension_tags:
                    raise headerExtensionError(f"'dim_{ddx}' must be a defined tag.")

                if f"dim_{ddx}_info" in json_dict\
                        and not isinstance(json_dict[f"dim_{ddx}_info"], str):
                    raise headerExtensionError(f"'dim_{ddx}_info' must be a string.")

                if f"dim_{ddx}_header" in json_dict\
                        and not isinstance(json_dict[f"dim_{ddx}_header"], dict):
                    raise headerExtensionError(f"'dim_{ddx}_header' must be a dict.")
            else:
                raise headerExtensionError(f" With {data_dimensions} dimensions the header extension"
                                           f" must contain 'dim_{ddx}'.")

    # 4. Check standard-defined data types
    for key in json_dict:
        if key in standard_defined and not check_type(json_dict[key], standard_defined[key][0]):
            raise headerExtensionError(f'{key} must be a {standard_defined[key][0]}. '
                                       f'{key} is a {type(json_dict[key])}, with value {json_dict[key]}.')

    # print('Header extension validated!')


def check_type(value, json_type):
    '''Checks that values is of type json_type
       json_type may be a tuple to handle array types
       e.g. (list, float) indicates a list of floats.
    '''
    if isinstance(json_type, tuple):
        while len(json_type) > 1:
            if not check_type(value, json_type[0]):
                return False
            try:
                # TO DO: check more than the first element!
                value = value[0]
            except TypeError:
                return False
            json_type = json_type[1:]
        return check_type(value, json_type[0])
    else:
        if isinstance(value, json_type):
            return True
    return False



"""Useful constants for use in mrs_tools functionality"""

# Gyromagnetic ratio in MHz/tesla
# From https://en.wikipedia.org/wiki/Gyromagnetic_ratio
# except for 1H https://physics.nist.gov/cgi-bin/cuu/Value?gammappbar
GYRO_MAG_RATIO = {
    '1H': 42.576,
    '2H': 6.536,
    '13C': 10.7084,
    '31P': 17.235}



""" Utility functions for NIfTI-MRS utilities

    Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
    Copyright (C) 2021 University of Oxford
"""

import numpy as np
from nibabel.nifti1 import Nifti1Extension


def modify_hdr_ext(new_hdr_ext, hdr):
    """Generate a new NIfTI header with a modified header extension.
    New header is a copy of the one passed

    :param new_hdr_ext: Modified header extension
    :type new_hdr_ext: dict
    :param hdr: NIfTI header
    :type hdr: nibabel.nifti2.Nifti2Header
    :return: Copied header with modified hdr extension
    :rtype: nibabel.nifti2.Nifti2Header
    """
    modded_hdr = hdr.copy()
    json_s = new_hdr_ext.to_json()
    extension = Nifti1Extension(44, json_s.encode('UTF-8'))
    modded_hdr.extensions.clear()
    modded_hdr.extensions.append(extension)

    return modded_hdr


def check_type(in_format):
    """Return type of header: long (list) or short (dict)

    :param in_format: Value of header key
    :type in_format: list or dict
    :return: 'long' or 'short'
    :rtype: str
    """
    if isinstance(in_format, list):
        return 'long'
    elif isinstance(in_format, dict)\
            and 'start' in in_format:
        return 'short'
    elif isinstance(in_format, dict)\
            and 'Value' in in_format:
        return check_type(in_format['Value'])


def dim_n_header_short_to_long(in_format, elements):
    if isinstance(in_format, list):
        return in_format
    elif isinstance(in_format, dict)\
            and 'start' in in_format:
        return _dict_to_list(in_format, elements)
    elif isinstance(in_format, dict)\
            and 'Value' in in_format:
        out = in_format.copy()
        out['Value'] = _dict_to_list(out['Value'], elements)
        return out


def dim_n_header_long_to_short(in_format):
    if isinstance(in_format, list):
        return _list_to_dict(in_format)
    elif isinstance(in_format, dict)\
            and 'start' in in_format:
        return in_format
    elif isinstance(in_format, dict)\
            and 'Value' in in_format:
        out = in_format.copy()
        out['Value'] = _list_to_dict(in_format['Value'])
        return out


def _dict_to_list(dict_desc, elements):
    """ Convert a short dict format to list format

    :param dict_desc: dict with start and increment fields
    :type dict_desc: dict
    :param elements: Number of elements in dimension
    :type elements: int
    :return: Converted list representation
    :rtype: list
    """
    if isinstance(dict_desc, dict):
        stop = dict_desc['start'] + dict_desc['increment'] * (elements - 1)
        return np.linspace(dict_desc['start'], stop, elements).tolist()
    else:
        return dict_desc


def _list_to_dict(list_desc):
    """If possible (i.e. vector is monotonic and equally spaced),
    convert list to dict (start + increment) representation.

    If conversion is not possible return the passed list.

    :param list_desc: List of header indicies
    :type list_desc: list
    :return: Dict representation or unmodified list
    :rtype: Dict or list
    """
    list_desc = np.asarray(list_desc)
    if np.issubdtype(list_desc.dtype, np.number)\
            and np.unique(np.diff(list_desc)).size == 1:
        return {'start': list_desc[0], 'increment': np.diff(list_desc)[0]}
    else:
        return list_desc.tolist()



"""Tools for merging, splitting and reordering the dimensions of NIfTI-MRS

    Author: Will Clarke <william.clarke@ndcn.ox.ac.uk>
    Copyright (C) 2021 University of Oxford
"""
import re

import numpy as np

from nifti_mrs.nifti_mrs import NIFTI_MRS, NIFTIMRS_DimDoesntExist
from nifti_mrs.tools import utils


class NIfTI_MRSIncompatible(Exception):
    pass


def split(nmrs, dimension, index_or_indicies):
    """Splits, or extracts indices from, a specified dimension of a
    NIFTI_MRS object. Output is two NIFTI_MRS objects. Header information preserved.

    :param nmrs: Input nifti_mrs object to split
    :type nmrs: fsl_mrs.core.nifti_mrs.NIFTI_MRS
    :param dimension: Dimension along which to split.
        Dimension tag or one of 4, 5, 6 (for 0-indexed 5th, 6th, and 7th)
    :type dimension: str or int
    :param index_or_indicies: Single integer index to split after,
        or list of interger indices to insert into second array.
        E.g. '0' will place the first index into the first output
        and 1 -> N in the second.
        '[1, 5, 10]' will place 1, 5 and 10 into the second output
        and all other will remain in the first.
    :type index_or_indicies: int or [int]
    :return: Two NIFTI_MRS object containing the split files
    :rtype: fsl_mrs.core.nifti_mrs.NIFTI_MRS
    """

    if isinstance(dimension, str):
        try:
            dim_index = nmrs.dim_position(dimension)
        except NIFTIMRS_DimDoesntExist:
            raise ValueError(f'{dimension} not found as dimension tag. This data contains {nmrs.dim_tags}.')
    elif isinstance(dimension, int):
        if dimension > (nmrs.ndim - 1) or dimension < 4:
            raise ValueError('Dimension must be one of 4, 5, or 6 (or DIM_TAG string).'
                             f' This data has {nmrs.ndim} dimensions,'
                             f' i.e. a maximum dimension value of {nmrs.ndim-1}.')
        dim_index = dimension
    else:
        raise TypeError('Dimension must be an int (4, 5, or 6) or string (DIM_TAG string).')

    # Construct indexing
    if isinstance(index_or_indicies, int):
        if index_or_indicies < 0\
                or (index_or_indicies + 1) >= nmrs.shape[dim_index]:
            raise ValueError('index_or_indicies must be between 0 and N-1,'
                             f' where N is the size of the specified dimension ({nmrs.shape[dim_index]}).')
        index = np.arange(index_or_indicies + 1, nmrs.shape[dim_index])

    elif isinstance(index_or_indicies, list):
        if not np.logical_and(np.asarray(index_or_indicies) >= 0,
                              np.asarray(index_or_indicies) <= nmrs.shape[dim_index]).all():
            raise ValueError('index_or_indicies must have elements between 0 and N,'
                             f' where N is the size of the specified dimension ({nmrs.shape[dim_index]}).')
        index = index_or_indicies

    else:
        raise TypeError('index_or_indicies must be single index or list of indicies')

    # Split header down
    split_hdr_ext_1, split_hdr_ext_2 = _split_dim_header(nmrs.hdr_ext,
                                                         dim_index + 1,
                                                         nmrs.shape[dim_index],
                                                         index_or_indicies)
    out_hdr_1 = utils.modify_hdr_ext(split_hdr_ext_1, nmrs.header)
    out_hdr_2 = utils.modify_hdr_ext(split_hdr_ext_2, nmrs.header)

    nmrs_1 = NIFTI_MRS(np.delete(nmrs[:], index, axis=dim_index), header=out_hdr_1)
    nmrs_2 = NIFTI_MRS(np.take(nmrs[:], index, axis=dim_index), header=out_hdr_2)

    return nmrs_1, nmrs_2


def _split_dim_header(hdr, dimension, dim_length, index):
    """Split dim_N_header keys in header extensions.

    :param hdr: Header extension to split
    :type hdr: dict
    :param dimension: Dimension (5, 6, or 7) to split along
    :type dimension: int
    :param dim_length: Length of dimension
    :type index: int
    :param index: Index to split after or indicies to extract
    :type index: int or list of ints
    :return: Split header eextension dicts
    :rtype: dict
    """
    hdr1 = hdr.copy()
    hdr2 = hdr.copy()

    def split_list(in_list):
        if isinstance(index, int):
            out_1 = in_list[:(index + 1)]
            out_2 = in_list[(index + 1):]
        elif isinstance(index, list):
            out_1 = np.delete(np.asarray(in_list), index).tolist()
            out_2 = np.take(np.asarray(in_list), index).tolist()
        return out_1, out_2

    def split_user_or_std(hdr_val):
        if isinstance(hdr_val, dict)\
                and 'Value' in hdr_val:
            tmp_1, tmp_2 = split_list(hdr_val['Value'])
            out_1 = hdr_val.copy()
            out_2 = hdr_val.copy()
            out_1.update({'Value': tmp_1})
            out_2.update({'Value': tmp_2})
            return out_1, out_2
        else:
            return split_list(hdr_val)

    def split_single(hdr_val):
        hdr_type = utils.check_type(hdr_val)
        long_fmt = utils.dim_n_header_short_to_long(hdr_val, dim_length)
        long_fmt_1, long_fmt_2 = split_user_or_std(long_fmt)
        if hdr_type == 'long':
            return long_fmt_1, long_fmt_2
        else:
            return utils.dim_n_header_long_to_short(long_fmt_1), utils.dim_n_header_long_to_short(long_fmt_2)

    key_str = f'dim_{dimension}_header'
    key_str_tag = f'dim_{dimension}'
    key_str_info = f'dim_{dimension}_info'
    if key_str in hdr:
        new_h1 = {}
        new_h2 = {}
        for sub_key in hdr[key_str]:
            new_h1[sub_key], new_h2[sub_key] = split_single(hdr[key_str][sub_key])

        curr_tag = hdr[key_str_tag]
        if key_str_info in hdr:
            curr_info = hdr[key_str_info]
        else:
            curr_info = None
        hdr1.set_dim_info(dimension - 5, curr_tag, info=curr_info, hdr=new_h1)
        hdr2.set_dim_info(dimension - 5, curr_tag, info=curr_info, hdr=new_h2)

    return hdr1, hdr2


def merge(array_of_nmrs, dimension):
    """Concatenate NIfTI-MRS objects along specified higher dimension

    :param array_of_nmrs: Array of NIFTI-MRS objects to concatenate
    :type array_of_nmrs: tuple or list of fsl_mrs.core.nifti_mrs.NIFTI_MRS
    :param dimension: Dimension along which to concatenate.
        Dimension tag or one of 4, 5, 6 (for 0-indexed 5th, 6th, and 7th).
    :type dimension: int or str
    :return: Concatenated NIFTI-MRS object
    :rtype: fsl_mrs.core.nifti_mrs.NIFTI_MRS
    """
    if isinstance(dimension, str):
        try:
            dim_index = array_of_nmrs[0].dim_position(dimension)
        except NIFTIMRS_DimDoesntExist:
            raise ValueError(f'{dimension} not found as dimension tag. This data contains {array_of_nmrs[0].dim_tags}.')
    elif isinstance(dimension, int):
        if dimension > (array_of_nmrs[0].ndim - 1) or dimension < 4:
            raise ValueError('Dimension must be one of 4, 5, or 6 (or DIM_TAG string).'
                             f' This data has {array_of_nmrs[0].ndim} dimensions,'
                             f' i.e. a maximum dimension value of {array_of_nmrs[0].ndim-1}.')
        dim_index = dimension
    else:
        raise TypeError('Dimension must be an int (4, 5, or 6) or string (DIM_TAG string).')

    # Check shapes and tags are compatible.
    # If they are and enter the data into a tuple for concatenation
    def check_shape(to_compare):
        for dim in range(to_compare.ndim):
            # Do not compare on selected dimension
            if dim == dim_index:
                continue
            if to_compare.shape[dim] != array_of_nmrs[0].shape[dim]:
                return False
        return True

    def check_tag(to_compare):
        for tdx in range(3):
            if array_of_nmrs[0].dim_tags[tdx] != to_compare.dim_tags[tdx]:
                return False
        return True

    to_concat = []
    for idx, nmrs in enumerate(array_of_nmrs):
        # Check shape
        if not check_shape(nmrs):
            raise NIfTI_MRSIncompatible('The shape of all concatentated objects must match.'
                                        f' The shape ({nmrs.shape}) of the {idx} object does'
                                        f' not match that of the first ({array_of_nmrs[0].shape}).')
        # Check dim tags for compatibility
        if not check_tag(nmrs):
            raise NIfTI_MRSIncompatible('The tags of all concatentated objects must match.'
                                        f' The tags ({nmrs.dim_tags}) of the {idx} object does'
                                        f' not match that of the first ({array_of_nmrs[0].dim_tags}).')

        if nmrs.ndim == dim_index:
            # If a squeezed singleton on the end.
            to_concat.append(np.expand_dims(nmrs[:], -1))
        else:
            to_concat.append(nmrs[:])

        # Merge header extension
        if idx == 0:
            merged_hdr_ext = nmrs.hdr_ext
            merged_length = to_concat[-1].shape[dim_index]
        else:
            merged_hdr_ext = _merge_dim_header(merged_hdr_ext,
                                               nmrs.hdr_ext,
                                               dim_index + 1,
                                               merged_length,
                                               to_concat[-1].shape[dim_index])
            merged_length += to_concat[-1].shape[dim_index]

    out_hdr = utils.modify_hdr_ext(merged_hdr_ext, array_of_nmrs[0].header)

    return NIFTI_MRS(np.concatenate(to_concat, axis=dim_index), header=out_hdr)


def _merge_dim_header(hdr1, hdr2, dimension, dim_length1, dim_length2):
    """Merge dim_N_header keys in header extensions.
    Output header copies all other fields from hdr1

    :param hdr1: header extension from 1st file
    :type hdr1: dict
    :param hdr2: header extension from 2nd file
    :type hdr2: dict
    :param dimension: Dimension (5,6, or 7) to merge along
    :type dimension: int
    :param dim_length1: Dimension length of first file
    :type dimension: int
    :param dim_length2: Dimension length of second file
    :type dimension: int
    :return: Merged header extension dict
    :rtype: dict
    """
    out_hdr = hdr1.copy()

    def merge_list(list_1, list_2):
        return list_1 + list_2

    def merge_user_or_std(hdr_val1, hdr_val2):
        if isinstance(hdr_val1, dict)\
                and 'Value' in hdr_val1:
            tmp = merge_list(hdr_val1['Value'], hdr_val2['Value'])
            out = hdr_val1.copy()
            out.update({'Value': tmp})
            return out
        else:
            return merge_list(hdr_val1, hdr_val2)

    def merge_single(hdr_val1, hdr_val2):
        hdr_type = utils.check_type(hdr_val1)
        long_fmt_1 = utils.dim_n_header_short_to_long(hdr_val1, dim_length1)
        long_fmt_2 = utils.dim_n_header_short_to_long(hdr_val2, dim_length2)
        long_fmt = merge_user_or_std(long_fmt_1, long_fmt_2)
        if hdr_type == 'long':
            return long_fmt
        else:
            return utils.dim_n_header_long_to_short(long_fmt)

    key_str = f'dim_{dimension}_header'
    key_str_tag = f'dim_{dimension}'
    key_str_info = f'dim_{dimension}_info'

    def run_check():
        # Check all other dimension fields are consistent
        dim_n = re.compile(r'dim_[567].*')
        for key in hdr1:
            if dim_n.match(key) and key != key_str:
                if hdr1[key] != hdr2[key]:
                    raise NIfTI_MRSIncompatible(f'Both files must have matching dimension headers apart from the '
                                                f'one being merged. {key} does not match.')

    if key_str in hdr1 and key_str in hdr2:
        run_check()
        # Check the subfields of the header to merge are consistent
        if not hdr1[key_str].keys() == hdr2[key_str].keys():
            raise NIfTI_MRSIncompatible(f'Both NIfTI-MRS files must have matching dim {dimension} header fields.'
                                        f'The first header contains {hdr1[key_str].keys()}. '
                                        f'The second header contains {hdr2[key_str].keys()}.')
        new_h = {}
        for sub_key in hdr1[key_str]:
            new_h[sub_key] = merge_single(hdr1[key_str][sub_key], hdr2[key_str][sub_key])

        curr_tag = hdr1[key_str_tag]
        if key_str_info in hdr1:
            curr_info = hdr1[key_str_info]
        else:
            curr_info = None
        out_hdr.set_dim_info(dimension - 5, curr_tag, info=curr_info, hdr=new_h)
    elif key_str in hdr1 and key_str not in hdr2\
            or key_str not in hdr1 and key_str in hdr2:
        # Incompatible headers
        raise NIfTI_MRSIncompatible(f'Both NIfTI-MRS files must have matching dim {dimension} header fields')
    elif key_str not in hdr1 and key_str not in hdr2:
        # Nothing to merge - still run check
        run_check()
    return out_hdr


def reorder(nmrs, dim_tag_list):
    """Reorder the higher dimensions of a NIfTI-MRS object.
    Can force a singleton dimension with new tag.

    :param nmrs: NIFTI-MRS object to reorder.
    :type nmrs: fsl_mrs.core.nifti_mrs.NIFTI_MRS
    :param dim_tag_list: List of dimension tags in desired order
    :type dim_tag_list: List of str
    :return: Reordered NIfTI-MRS object.
    :rtype: fsl_mrs.core.nifti_mrs.NIFTI_MRS
    """

    # Check existing tags are in the list of desired tags
    for idx, tag in enumerate(nmrs.dim_tags):
        if tag not in dim_tag_list\
                and tag is not None:
            raise NIfTI_MRSIncompatible(
                f'The existing tag ({tag}) does not appear '
                f'in the requested tag order ({dim_tag_list}).')

    # Create singleton dimensions if required
    original_dims = nmrs.ndim
    new_dim = sum(x is not None for x in nmrs.dim_tags) + 4
    dims_to_add = tuple(range(original_dims, new_dim + 1))
    data_with_singleton = np.expand_dims(nmrs[:], dims_to_add)

    # Create list of source indicies
    # Create list of destination indicies
    # Keep track of singleton tags
    source_indicies = []
    dest_indicies = []
    singleton_tags = {}
    counter = 0
    for idx, tag in enumerate(dim_tag_list):
        if tag is not None:
            if tag in nmrs.dim_tags:
                source_indicies.append(nmrs.dim_tags.index(tag) + 4)
            else:
                source_indicies.append(nmrs.ndim + counter)
                counter += 1
                singleton_tags.update({(idx + 5): tag})

            dest_indicies.append(idx + 4)

    # Sort header extension dim_tags
    dim_n = re.compile(r'dim_[567].*')
    new_hdr_ext = nmrs.hdr_ext.copy()
    new_hdr_ext.dimensions = data_with_singleton.ndim
    for key in nmrs.hdr_ext:
        if dim_n.match(key):
            new_index = dest_indicies[source_indicies.index(int(key[4]) - 1)] + 1
            new_ind_str = f'{new_index}th'
            new_hdr_ext.set_dim_info(new_ind_str, nmrs.hdr_ext[key])

    # For any singleton dimensions we've added
    for dim in singleton_tags:
        new_hdr_ext.set_dim_info(f'{dim}th', singleton_tags[dim])

    new_header = nmrs.header.copy()
    new_nmrs = NIFTI_MRS(
        np.moveaxis(data_with_singleton, source_indicies, dest_indicies),
        header=new_header)

    # Finaly insert new header extension
    new_nmrs.hdr_ext = new_hdr_ext
    return new_nmrs



'''Definitions of NIfTI-MRS standard meta data and dimension tags.

Type fields should either be generic python types: float, int, str
or a tuple indicating an array type and element type : (list, float) or (list, str)

Copyright Will Clarke, University of Oxford, 2021
'''

# Define nifti-mrs version number here.
# First element is major version, second is minor
nifti_mrs_version = [0, 6]

# Possible dimension tags and descriptions
dimension_tags = {"DIM_COIL": "For storage of data from each individual receiver coil element.",
                  "DIM_DYN": "For storage of each individual acquisition transient. E.g. for post-acquisition B0 drift correction.",
                  "DIM_INDIRECT_0": "The indirect detection dimension - necessary for 2D (and greater) MRS acquisitions.",
                  "DIM_INDIRECT_1": "The indirect detection dimension - necessary for 2D (and greater) MRS acquisitions.",
                  "DIM_INDIRECT_2": "The indirect detection dimension - necessary for 2D (and greater) MRS acquisitions.",
                  "DIM_PHASE_CYCLE": "Used for the time-proportional phase incrementation method.",
                  "DIM_EDIT": "Used for edited MRS techniques such as MEGA or HERMES.",
                  "DIM_MEAS": "Used to indicate multiple repeats of the full sequence contained within the same original data file.",
                  "DIM_USER_0": "User defined dimension.",
                  "DIM_USER_1": "User defined dimension.",
                  "DIM_USER_2": "User defined dimension.",
                  "DIM_ISIS": "Dimension for storing image-selected in vivo spectroscopy (ISIS) acquisitions."}

# Required metadata fields
required = {'SpectrometerFrequency':
            {'doc': 'Precession frequency in MHz of the nucleus being addressed for each spectral axis.',
             'type': (list, float)},
            'ResonantNucleus':
            {'doc': 'Must be one of the DICOM recognised nuclei “1H”, “3HE”, “7LI”, “13C”, “19F”, “23NA”, “31P”, “129XE” or one named in the specified format. I.e. Mass number followed by the chemical symbol in uppercase.',
             'type': (list, str)}}

# Defined metadata fields
# # 5.1 MRS specific Tags
# 'EchoTime'
# 'RepetitionTime'
# 'InversionTime'
# 'MixingTime'
# 'AcquisitionStartTime'
# 'ExcitationFlipAngle'
# 'TxOffset'
# 'VOI'
# 'WaterSuppressed'
# 'WaterSuppressionType'
# 'SequenceTriggered'
# # 5.2 Scanner information
# 'Manufacturer'
# 'ManufacturersModelName'
# 'DeviceSerialNumber'
# 'SoftwareVersions'
# 'InstitutionName'
# 'InstitutionAddress'
# 'TxCoil'
# 'RxCoil'
# # 5.3 Sequence information
# 'SequenceName'
# 'ProtocolName'
# # 5.4 Sequence information
# 'PatientPosition'
# 'PatientName'
# 'PatientID'
# 'PatientWeight'
# 'PatientDoB'
# 'PatientSex'
# # 5.5 Provenance and conversion metadata
# 'ConversionMethod'
# 'ConversionTime'
# 'OriginalFile'
# # 5.6 Spatial information
# 'kSpace'

# These fields are optional but must not be redefined.
# Format is a dict of tuples containing (type, unit string, doc string, anonymisation state)
standard_defined = {
    # 5.1 MRS specific Tags
    'EchoTime':
        (float,
         's',
         'Time from centroid of excitation to start of FID or centre of echo. Units: Seconds',
         False),
    'RepetitionTime':
        (float,
         's',
         'Sequence repetition time. Units: Seconds',
         False),
    'InversionTime':
        (float,
         's',
         'Inversion time. Units: Seconds',
         False),
    'MixingTime':
        (float,
         's',
         'Mixing time in e.g. STEAM sequence. Units: Seconds',
         False),
    'AcquisitionStartTime':
        (float,
         's',
         'Time, relative to EchoTime, that the acquisition starts. Positive values indicate a time after the EchoTime, negative indicate before the EchoTime, a value of zero indicates no offset. Units: Seconds',
         False),
    'ExcitationFlipAngle':
        (float,
         'degrees',
         'Nominal excitation pulse flip-angle',
         False),
    'TxOffset':
        (float,
         'ppm',
         'Transmit chemical shift offset from SpectrometerFrequency',
         False),
    'VOI':
        ((list, list, float),
         None,
         'VoI localisation volume for MRSI sequences. Stored as a 4 x 4 affine using identical conventions to the xform NIfTI affine matrix. Not defined for data stored with a single spatial voxel',
         False),
    'WaterSuppressed':
        (bool,
         None,
         'Boolian value indicating whether data was collected with (True) or without (False) water suppression.',
         False),
    'WaterSuppressionType':
        (str,
         None,
         'Type of water suppression used.',
         False),
    'SequenceTriggered':
        (bool,
         None,
         'Boolian value indicating whether the sequence is triggered. If triggered the repetition time might not be constant.',
         False),
    # 5.2 Scanner information
    'Manufacturer':
        (str,
         None,
         'Manufacturer of the device. DICOM tag (0008,0070).',
         False),
    'ManufacturersModelName':
        (str,
         None,
         "Manufacturer's model name of the device. DICOM tag (0008,1090).",
         False),
    'DeviceSerialNumber':
        (str,
         None,
         "Manufacturer's serial number of the device. DICOM tag (0018,1000).",
         False),
    'SoftwareVersions':
        (str,
         None,
         "Manufacturer's designation of the software version. DICOM tag (0018,1020)",
         False),
    'InstitutionName':
        (str,
         None,
         "Institution’s Name. DICOM tag (0008,0080).",
         False),
    'InstitutionAddress':
        (str,
         None,
         "Institution’s address. DICOM tag (0008,0081).",
         False),
    'TxCoil':
        (str,
         None,
         "Name or description of transmit RF coil.",
         False),
    'RxCoil':
        (str,
         None,
         "Name or description of receive RF coil.",
         False),
    # 5.3 Sequence information
    'SequenceName':
        (str,
         None,
         "User defined name. DICOM tag (0018,0024).",
         False),
    'ProtocolName':
        (str,
         None,
         "User-defined description of the conditions under which the Series was performed. DICOM tag (0018,1030).",
         False),
    # 5.4 Sequence information
    'PatientPosition':
        (str,
         None,
         "Patient position descriptor relative to the equipment. DICOM tag (0018,5100). Must be one of the DICOM defined code strings e.g. HFS, HFP.",
         False),
    'PatientName':
        (str,
         None,
         "Patient's full name. DICOM tag (0010,0010).",
         True),
    'PatientID':
        (str,
         None,
         "Patient identifier. DICOM tag (0010,0020).",
         True),
    'PatientWeight':
        (float,
         'kg',
         "Weight of the Patient in kilograms. DICOM tag (0010,1030).",
         False),
    'PatientDoB':
        (str,
         None,
         "Date of birth of the named Patient. YYYYMMDD. DICOM tag (0010,0030).",
         True),
    'PatientSex':
        (str,
         None,
         "Sex of the named Patient. ‘M’, ‘F’, ‘O’. DICOM tag (0010,0040)",
         False),
    # 5.5 Provenance and conversion metadata
    'ConversionMethod':
        (str,
         None,
         "Description of the process or program used for conversion. May include additional information like software version.",
         False),
    'ConversionTime':
        (str,
         None,
         "Time and date of conversion. ISO 8601 compliant format",
         False),
    'OriginalFile':
        ((list, str),
         None,
         "Name and extension of the original file(s)",
         True),
    # 5.6 Spatial information
    'kSpace':
        ((list, bool),
         None,
         "Three element list, corresponding to the first three spatial dimensions. If True the data is stored as a dense k-space representation.",
         False)}
